import { Duck } from '../..';
import Game from '../game';
import Sprite from '../gameobjects/sprite';
import Scene from '../scene';
import Animation from './animation';
import StateMachine from './stateMachine';
export default class AnimationManager {
    /**
     * @memberof AnimationManager
     * @description Game instance
     * @type Game
     * @since 2.0.0
     */
    game: Game;
    /**
     * @memberof AnimationManager
     * @description Scene instance
     * @type Scene
     * @since 2.0.0
     */
    scene: Scene;
    /**
     * @memberof AnimationManager
     * @description Sprite that the AnimationManager is attached to
     * @type Sprite
     * @since 2.0.0
     */
    sprite: Sprite;
    /**
     * @memberof AnimationManager
     * @description Animations that are registered and added to the AnimationManager
     * @type Animation[]
     * @since 2.0.0
     */
    animations: Animation[];
    /**
     * @memberof AnimationManager
     * @description The current selected animation, last played or currently playing
     * @type Animation
     * @since 2.0.0
     */
    currentAnimation: Animation;
    /**
     * @constructor AnimationManager
     * @param {Game} game Game instance
     * @param {Scene} scene Scene instance
     * @param {Sprite} sprite Sprite that the AnimationManager is attached to
     * @param {Duck.Types.Animation.Config} defaultAnimation Configuration to make a default animation of the Sprite
     * @since 2.0.0
     */
    constructor(game: Game, scene: Scene, sprite: Sprite, defaultAnimation: Duck.Types.Animation.Config);
    /**
     * @memberof AnimationManager
     * @description Creates a StateMachine
     * @param {Duck.Types.StateMachine.Config} config StateMachine config
     * @param {Animation[]} animations Animations to add to the StateMachine
     * @returns {StateMachine}
     * @since 2.0.0
     */
    createStateMachine(config: Duck.Types.StateMachine.Config, animations: Animation[]): StateMachine;
    /**
     * @memberof AnimationManager
     * @description Sets the current animation
     * @param {string} key Key of the Animation to set as the currentAnimation
     * @since 2.0.0
     */
    setCurrentAnimation(key: string): void;
    /**
     * @memberof AnimationManager
     * @description Adds and creates a new Animation
     * @param {string} key Key of the Animation to set as the currentAnimation
     * @returns {Animation}
     * @since 2.0.0
     */
    add(config: Duck.Types.Animation.Config): Animation;
    /**
     * @memberof AnimationManager
     * @description Removes an Animation from the array of animations (AnimationManager#animations)
     * @param {string} key Key of the Animation to remove
     * @since 2.0.0
     */
    remove(key: string): void;
    /**
     * @memberof AnimationManager
     * @description Finds and returns an Animation from the array of animations (AnimationManager#animations)
     * @param {string} key Key of the Animation find
     * @returns {Animation | undefined}
     * @since 2.0.0
     */
    get(key: string): Animation | undefined;
    /**
     * @memberof AnimationManager
     * @description Finds and returns the index of an Animation from the array of animations (AnimationManager#animations)
     * @param {string} key Key of the Animation find the index of
     * @returns {number}
     * @since 2.0.0
     */
    getIndex(key: string): number;
    /**
     * @memberof AnimationManager
     * @description Clears all animations
     * @since 2.0.0
     */
    clear(): void;
    /**
     * @memberof AnimationManager
     * @description Sets the current animation repeat
     * @param {number} repeat
     * @since 2.0.0
     */
    setRepeat(repeat: number): void;
    /**
     * @memberof AnimationManager
     * @description Sets the current animation reversed repeat
     * @param {number} repeat
     * @since 2.0.0
     */
    setRepeatReverse(repeat: number): void;
    /**
     * @memberof AnimationManager
     * @description Plays the current animation normally
     * @since 2.0.0
     */
    play(): void;
    /**
     * @memberof AnimationManager
     * @description Plays the current animation reversely
     * @since 2.0.0
     */
    playReverse(): void;
    /**
     * @memberof AnimationManager
     * @description Pauses the current animation normally
     * @since 2.0.0
     */
    pause(): void;
    /**
     * @memberof AnimationManager
     * @description Pauses the current animation reversely
     * @since 2.0.0
     */
    pauseReverse(): void;
    /**
     * @memberof AnimationManager
     * @description Resumes the current animation normally
     * @since 2.0.0
     */
    resume(): void;
    /**
     * @memberof AnimationManager
     * @description Resumes the current animation reversely
     * @since 2.0.0
     */
    resumeReverse(): void;
    /**
     * @memberof AnimationManager
     * @description Stops the current animation normally
     * @since 2.0.0
     */
    stop(): void;
    /**
     * @memberof AnimationManager
     * @description Stops the current animation reversely
     * @since 2.0.0
     */
    stopReverse(): void;
    /**
     * @memberof AnimationManager
     * @description Restarts the current animation normally
     * @since 2.0.0
     */
    restart(): void;
    /**
     * @memberof AnimationManager
     * @description Restarts the current animation reversely
     * @since 2.0.0
     */
    restartReverse(): void;
}
