import { Duck } from '../..';
import Game from '../game';
import Group from '../group/group';
import Vector2 from '../math/vector2';
import Scene from '../scene';
import Collider from './models/collider';
import Hitbox from './models/hitbox';
/**
 * @class PhysicsBody
 * @classdesc Creates a DuckEngine PhysicsBody
 * @description The PhysicsBody Class. The GameObject class extends this class
 * @since 2.0.0
 */
export default class PhysicsBody<textureType extends Duck.Types.Texture.Type> {
    /**
     * @memberof PhysicsBody
     * @description The unique identifier for a GameObject
     * @type string
     * @since 2.0.0
     */
    readonly id: string;
    /**
     * @memberof PhysicsBody
     * @description The shape of the GameObject, 'rect', 'circle', 'roundrect', or 'sprite'
     * @type Duck.Types.Collider.ShapeString
     * @since 2.0.0
     */
    readonly shape: Duck.Types.Collider.ShapeString;
    /**
     * @memberof PhysicsBody
     * @description The current global position of the GameObject
     * @type Vector2
     * @since 2.0.0
     */
    position: Vector2;
    /**
     * @memberof PhysicsBody
     * @description The width of the GameObject
     * @type number
     * @since 2.0.0
     */
    w: number;
    /**
     * @memberof PhysicsBody
     * @description The height of the GameObject
     * @type number
     * @since 2.0.0
     */
    h: number;
    /**
     * @memberof PhysicsBody
     * @description The radius of the GameObject
     * @type number
     * @since 2.0.0
     */
    r: number;
    /**
     * @memberof PhysicsBody
     * @description PhysicsBody config, includes: type - KinematicBody | RigidBody | StaticBody
     *
     * defaults: { type: 'KinematicBody'}
     *
     * @type Duck.Types.PhysicsBody.Config
     * @since 2.0.0
     */
    options: Duck.Types.PhysicsBody.Config;
    /**
     * @memberof PhysicsBody
     * @description The Game instance
     * @type Game
     * @since 2.0.0
     */
    game: Game;
    /**
     * @memberof PhysicsBody
     * @description The Scene instance
     * @type Game
     * @since 2.0.0
     */
    scene: Scene;
    /**
     * @memberof PhysicsBody
     * @description The Collider instance of the PhysicsBody
     * @type Collider | undefined
     * @since 2.0.0
     */
    collider: Collider | undefined;
    /**
     * @memberof PhysicsBody
     * @description An array or group of GameObjects that can collide with the PhysicsBody
     * @type Duck.TypeClasses.GameObjects.GameObject<textureType>[] | Group<Duck.TypeClasses.GameObjects.GameObject<textureType>>
     * @since 2.0.0
     */
    collidesWith: Duck.TypeClasses.GameObjects.GameObject<textureType>[] | Group<Duck.TypeClasses.GameObjects.GameObject<textureType>>;
    /**
     * @memberof PhysicsBody
     * @description The Collider Hitbox of the PhysicsBody
     * @type Hitbox | undefined
     * @since 2.0.0
     */
    hitbox: Hitbox | undefined;
    /**
     * @memberof PhysicsBody
     * @description The velocity of the PhysicsBody
     * @type Vector2
     * @since 2.0.0
     */
    velocity: Vector2;
    /**
     * @memberof PhysicsBody
     * @description The bounds of the PhysicsBody
     * @type Duck.Types.Math.BoundsLike
     * @since 2.0.0
     */
    bounds: Duck.Types.Math.BoundsLike;
    /**
     * @memberof PhysicsBody
     * @description Determines if the PhysicsBody._update is called by the Scene.physicsServer used by Scene.physicsList
     * , changing this value does nothing, must use PhysicsBody.setEnabled
     * @type boolean
     * @since 2.0.0
     */
    enabled: boolean;
    /**
     * @memberof PhysicsBody
     * @description Determines if the PhysicsBody is attached to another PhysicsBody
     * @type boolean
     * @since 2.0.0
     */
    isAttached: boolean;
    /**
     * @memberof PhysicsBody
     * @description PhysicsBodies that are attached
     * @type PhysicsBody<Duck.Types.Texture.Type>[]
     * @since 2.0.0
     */
    attachedChildren: PhysicsBody<Duck.Types.Texture.Type>[];
    /**
     * @memberof PhysicsBody
     * @description The offset between the PhysicsBody that self is attached to
     * @type Vector2
     * @since 2.0.0
     */
    attachOffset: Vector2;
    /**
     * @memberof PhysicsBody
     * @description Object that has all the physics method
     * @type { addCollider: (collidesWith: Duck.TypeClasses.GameObjects.GameObject[]) => Collider;
     *	setBounds: (x: number, y: number, w: number, h: number) => void
     * }
     * @since 2.0.0
     */
    physics: {
        /**
         * @memberof PhysicsBody#physics
         * @description Adds a collider to the PhysicsBody
         * @param {Duck.TypeClasses.GameObjects.GameObject<textureType>[]} collidesWith What the GameObject collides with
         * @since 2.0.0
         */
        addCollider: (collidesWith: Duck.TypeClasses.GameObjects.GameObject<textureType>[]) => Collider | undefined;
        /**
         * @memberof PhysicsBody#physics
         * @description Adds a hitbox to the PhysicsBody
         * @param {number} [w] Width of hitbox, optional -> defaults: PhysicsBody.w or PhysicsBody.r * 2
         * @param {number} [h] Height of hitbox, optional -> defaults: PhysicsBody.h or PhysicsBody.r * 2
         * @param {Vector2} [offset=Vector2.ZERO] Offset of hitbox, optional -> defaults: Vector2.ZERO
         * @since 2.0.0
         */
        addHitbox: (w?: number, h?: number, offset?: Vector2) => Hitbox;
        /**
         * @memberof GameObject#physics
         * @description Adds bounds to the GameObject
         * @param {number} x X position
         * @param {number} y Y position
         * @param {number} w Width of the bounds
         * @param {number} h Height of the bounds
         * @since 2.0.0
         */
        setBounds: (x: number, y: number, w: number, h: number) => void;
    };
    /**
     * @constructor PhysicsBody
     * @description Creates a PhysicsBody instance. Extended by GameObject
     * @param {Duck.Types.Collider.ShapeString} shape Shape of PhysicsBody
     * @param {number} id ID from GameObject ID
     * @param {number} x X position
     * @param {number} y Y position
     * @param {number} w Width
     * @param {number} h Height
     * @param {number} r Radius
     * @param {Game} game Game instance
     * @param {Scene} scene Scene instance
     * @since 2.0.0
     */
    constructor(shape: Duck.Types.Collider.ShapeString, id: string, x: number, y: number, w: number, h: number, r: number, game: Game, scene: Scene);
    setEnabled(enabled: boolean): boolean;
    /**
     * @memberof PhysicsBody
     * @description Updates the PhysicsBody's position by the velocity. Sets velocity to 0 on every tick.
     * Clamps position to bounds if exists. Rounds pixels if roundPixels game config is set to true.
     * Updates hitbox.collisionState if hitbox exists.
     *
     * DO NOT CALL MANUALLY, CALLED IN SCENE.__tick
     *
     * @since 2.0.0
     */
    _update(): void;
    /**
     * @memberof PhysicsBody
     * @description Sets the PhysicsBody type, PhysicsBody type determines what can be applied and what effects the body
     * @param {Duck.Types.PhysicsBody.Type} type The type of the PhysicsBody, 'KinematicBody' | 'RigidBody' | 'StaticBody'
     * @since 2.0.0
     */
    setType(type: Duck.Types.PhysicsBody.Type): Duck.Types.PhysicsBody.Config;
    /**
     * @memberof PhysicsBody
     * @description Attaches self to another PhysicsBody, makes self unmovable and follows the PhysicsBody accordingly
     * @param {PhysicsBody<Duck.Types.Texture.Type>} object PhysicsBody to attach to
     * @param {Vector2} [diffOffset=Vector2] A different offset, optional -> defaults: Difference in positions from PhysicsBody to self
     * @since 2.0.0
     */
    attachTo(object: PhysicsBody<Duck.Types.Texture.Type>, diffOffset?: Vector2): void;
    /**
     * @memberof PhysicsBody
     * @description Attaches a child PhysicsBody to self, makes child unmovable and follows the self accordingly
     * @param {PhysicsBody<Duck.Types.Texture.Type>} object PhysicsBody to attach
     * @param {Vector2} [diffOffset=Vector2] A different offset, optional -> defaults: Difference in positions from self to PhysicsBody
     * @since 2.0.0
     */
    attachChild(object: PhysicsBody<Duck.Types.Texture.Type>, diffOffset?: Vector2): void;
    /**
     * @memberof PhysicsBody
     * @description Detaches self from another PhysicsBody
     * @param {PhysicsBody<Duck.Types.Texture.Type>} object PhysicsBody to detach from
     * @since 2.0.0
     */
    detachFrom(object: PhysicsBody<Duck.Types.Texture.Type>): void;
    /**
     * @memberof PhysicsBody
     * @description Detaches PhysicsBody from self
     * @param {PhysicsBody<Duck.Types.Texture.Type>} object PhysicsBody to detach
     * @since 2.0.0
     */
    detachChild(object: PhysicsBody<Duck.Types.Texture.Type>): void;
    /**
     * @memberof PhysicsBody
     * @description Sets the velocity based on an axis, PhysicsBody.options.type must be KinematicBody
     * @param {'x'|'y'} axis The axis to set the velocity of
     * @param {number} pxPerSecond The value to set the velocity axis as, in pixels per second
     * @since 2.0.0
     */
    setVelocity(axis: 'x' | 'y', pxPerSecond: number): void;
    /**
     * @memberof PhysicsBody
     * @description Sets the velocity.x, PhysicsBody.options.type must be KinematicBody
     * @param {number} pxPerSecond The value to set the velocity axis as, in pixels per second
     * @since 2.0.0
     */
    setVelocityX(pxPerSecond: number): void;
    /**
     * @memberof PhysicsBody
     * @description Sets the velocity.y, PhysicsBody.options.type must be KinematicBody
     * @param {number} pxPerSecond The value to set the velocity.y as, in pixels per second
     * @since 2.0.0
     */
    setVelocityY(pxPerSecond: number): void;
    /**
     * @memberof PhysicsBody
     * @description Accelerates the velocity by an amount, PhysicsBody.options.type must be KinematicBody
     * @param {Vector2} target The target velocity
     * @param {number} amount The value to increase the velocity by
     * @since 2.0.0
     */
    accelerateVelocity(target: Vector2, amount: number): void;
    /**
     * @memberof PhysicsBody
     * @description Applies friction to the velocity by an amount, PhysicsBody.options.type must be KinematicBody or RigidBody
     * @param {number} frictionAmount The value to decrease the velocity by
     * @since 2.0.0
     */
    applyFriction(frictionAmount: Vector2): void;
    /**
     * @memberof PhysicsBody
     * @description Applies gravity to the velocity by a Vector2, PhysicsBody.options.type must be KinematicBody or RigidBody
     * @param {Vector2} gravity The Vector2 to add to the velocity by
     * @since 2.0.0
     */
    applyGravity(gravity: Vector2): void;
    /**
     * @memberof PhysicsBody
     * @description Applies gravity to the velocity by a Vector2, PhysicsBody.options.type must be KinematicBody or RigidBody
     * @param {Duck.Types.Math.BoundsLike} [bounds=PhysicsBody.bounds] The bounds of the PhysicsBody, optional -> defaults: PhysicsBody.bounds, if none
     * are set, it is infinite
     * @param {number} [restitution=1] How much energy is lost when bouncing, a number between 0-1 to loose energy,
     * 1-any to increase energy, 1 = none, must be a positive number
     * @since 2.0.0
     */
    bounceVelocityBounds(bounds?: Duck.Types.Math.BoundsLike, restitution?: number): void;
    /**
     * @memberof PhysicsBody
     * @description Reflects the velocity, sets the velocity as the opposite value of the velocity, PhysicsBody.options.type must be KinematicBody or RigidBody
     *
     * @example myPhysicsBody.setVelocity('x', 3);
     * myPhysicsBody.reflect(); // velocity: 0, -3
     *
     * @since 2.0.0
     */
    reflectVelocity(): void;
    /**
     * @memberof PhysicsBody
     * @description Auto scales the PhysicsBody.hitbox to fit the shape
     * @param {Vector2} [offset] Position offset, optional -> defaults: undefined
     * @since 2.0.0
     */
    autoFitHitbox(offset?: Vector2): void;
    /**
     * @memberof PhysicsBody
     * @description Scales the PhysicsBody.hitbox
     * @param {Vector2} scale Scale Vector2, x is width, y is height
     * @since 2.0.0
     */
    scaleHitbox(scale: Vector2): void;
    /**
     * @memberof PhysicsBody
     * @description Gets the top most coordinate of the PhysicsBody
     * @returns {number}
     * @since 2.0.0
     */
    getTop(): number;
    /**
     * @memberof PhysicsBody
     * @description Gets the bottom most coordinate of the PhysicsBody
     * @returns {number}
     * @since 2.0.0
     */
    getBottom(): number;
    /**
     * @memberof PhysicsBody
     * @description Gets the left most coordinate of the PhysicsBody
     * @returns {number}
     * @since 2.0.0
     */
    getLeft(): number;
    /**
     * @memberof PhysicsBody
     * @description Gets the right most coordinate of the PhysicsBody
     * @returns {number}
     * @since 2.0.0
     */
    getRight(): number;
    /**
     * @memberof PhysicsBody
     * @description Gets the center coordinates of the PhysicsBody
     * @returns {Vector2}
     * @since 2.0.0
     */
    getCenter(): Vector2;
    /**
     * @memberof PhysicsBody
     * @description Gets the centerY coordinate of the PhysicsBody
     * @returns {number}
     * @since 2.0.0
     */
    getCenterY(): number;
    /**
     * @memberof PhysicsBody
     * @description Gets the centerX coordinate of the PhysicsBody
     * @returns {number}
     * @since 2.0.0
     */
    getCenterX(): number;
    /**
     * @memberof PhysicsBody
     * @description Checks and returns the Collision Type if two hitboxes are colliding
     * @param {PhysicsBody<Duck.Types.Texture.Type>} obj PhysicsBody to check their hitbox with
     * @returns {false | Duck.Types.Collider.CollisionResponseType | undefined}
     * @since 2.0.0
     */
    isColliding(obj: PhysicsBody<Duck.Types.Texture.Type>): false | Duck.Types.Collider.CollisionResponseType | undefined;
    /**
     * @memberof PhysicsBody
     * @description Checks and returns the Collision Type if multiple hitboxes are colliding
     * @param {Group<PhysicsBody<Duck.Types.Texture.Type>> | PhysicsBody<Duck.Types.Texture.Type>[]} objects PhysicsBodies to check their hitbox with
     * @returns {false | Duck.Types.Collider.CollisionResponseType}
     * @since 2.0.0
     */
    isCollidingGroup(objects: Group<PhysicsBody<Duck.Types.Texture.Type>> | PhysicsBody<Duck.Types.Texture.Type>[]): false | "left" | "right" | "bottom" | "top";
}
